classdef EllipseFunction6_3d
    %FUNCLASS Summary of this class goes here
    %   Detailed explanation goes here
    
    properties
        Params = [0.2;
                  4;
                  4];
    end
    
    methods (Access = public)
        
        function out = Xscale(obj)
            out = obj.Params(1);
        end
        
        function out = Yscale(obj)
            out = obj.Params(2);
        end
        
        function out = Zscale(obj)
            out = obj.Params(3);
        end
        
        function pfixout = pfix(obj,h_max)
            pfixtmp =[obj.Xscale 0 0
                     -obj.Xscale 0 0];
            circ = 2*pi*obj.Yscale;
            no_extrapfixs = ceil(circ/h_max);
          
            extrapvec = [0:1:no_extrapfixs-1]';
            pfixtmp2 = [0*extrapvec sin(2*pi/no_extrapfixs*extrapvec)*obj.Yscale, cos(2*pi/no_extrapfixs*extrapvec)*obj.Zscale];
            pfixout = [pfixtmp;pfixtmp2];
            
            
            
            
%                 0 obj.Yscale 0
%                 0 -obj.Yscale 0
%                 0 0 obj.Zscale
%                 0 0 -obj.Zscale];
        end
        
        function pboundsout = pbounds(obj)
            pboundsout =[-obj.Xscale -obj.Yscale -obj.Zscale
                obj.Xscale  obj.Yscale  obj.Zscale];
        end
        
        function [Func] = fd(obj,p)
            %Func This returns the distance function
            % This returns signed distance function for a given point
            
            Func = ((p(:,1)./obj.Xscale).^2 + (p(:,2)./obj.Yscale).^2 + (p(:,3)./obj.Zscale).^2).^0.5 - 1;
%             Func = sign((p(:,1)./obj.Xscale).^2 + (p(:,2)./obj.Yscale).^2 + (p(:,3)./obj.Zscale).^2 - 1).*(sum(obj.distance(p).^2,2)).^0.5;
        end
        
        function [Func] = F(obj,p)
            %Func This returns the distance function
            % This returns signed distance function for a given point
            
            Func = ((p(:,1)./obj.Xscale).^2 + (p(:,2)./obj.Yscale).^2 + (p(:,3)./obj.Zscale).^2) - 1;
        end
        
        function [Grad] = dF(obj,p)
            %dF Summary of this method goes here
            %  This returnd the first order partial derivitves of the
            %  function for a given point
            dRdu1(:,1) = 2*(p(:,1)./obj.Xscale.^2);
            dRdu2(:,1) = 2*(p(:,2)./obj.Yscale.^2);
            dRdu3(:,1) = 2*(p(:,3)./obj.Zscale.^2);
            Grad = [dRdu1 dRdu2 dRdu3];
        end
        
        function [SecondGrad] = ddF(obj,p)
            %ddF Second order partial derivitives (risidual derivitives)
            %  used for La Grange and CB2 Ellim approx
            
            ddRdxdx = 2/obj.Xscale^2;
            ddRdxdy = 0;
            ddRdxdz = 0;
            ddRdydx = 0;
            ddRdydy = 2/obj.Yscale^2;
            ddRdydz = 0;
            ddRdzdx = 0;
            ddRdzdy = 0;
            ddRdzdz = 2/obj.Zscale^2;
            
            SecondGrad = [ddRdxdx ddRdxdy ddRdxdz;
                          ddRdydx ddRdydy ddRdydz;
                          ddRdzdx ddRdzdy ddRdzdz];
        end
        
        function [DF] = dFelimActual(obj,p,slavedof)
            if slavedof == 1
                s = p(1,1); %x
                m1 = p(1,2);
                m2 = p(1,3);
                
                as = obj.Xscale;
                bm1 = obj.Yscale;
                bm2 = obj.Zscale;
            elseif slavedof == 2
                s = p(1,2); %y slave
                m1 = p(1,1);
                m2 = p(1,3);
                
                as = obj.Yscale;
                bm1 = obj.Xscale;
                bm2 = obj.Zscale;
            elseif slavedof == 3
                s = p(1,3); %z slave
                m1 = p(1,1);
                m2 = p(1,2);
                
                as = obj.Zscale;
                bm1 = obj.Xscale;
                bm2 = obj.Yscale;
            end
            
            dSdm1 =-(as^2*m1)/(bm1^2*s);
            dSdm2 =-(as^2*m2)/(bm2^2*s);

            DF = [dSdm1 dSdm2];
        end
        
        function [SecondGrad] = ddFelimActual(obj,p,slavedof)
            %ddFelimActual Second order actual derivitives for ellimination
            %  used for La Grange and CB2 Ellim approx
            
            if slavedof == 1
                s = p(1,1); %x slave
                m1 = p(1,2);
                m2 = p(1,3);
                
                as = obj.Xscale;
                bm1 = obj.Yscale;
                bm2 = obj.Zscale;
                
            elseif slavedof == 2
                s = p(1,2); %y slave
                m1 = p(1,1);
                m2 = p(1,3);
                
                as = obj.Yscale;
                bm1 = obj.Xscale;
                bm2 = obj.Zscale;
            elseif slavedof == 3
                
                s = p(1,3); %z slave
                m1 = p(1,1);
                m2 = p(1,2);
                
                as = obj.Zscale;
                bm1 = obj.Xscale;
                bm2 = obj.Yscale;
            end
            
            % ddS1d1d1 = -(as/bm1)^2*(s.^2 + (as/bm1)^2*m1.^2)/s^3;
            % ddS1d1d2 = (as^4/(bm1^2*bm2^2))*m1*m2/s^3;
            % ddS2d2d1 = (as^4/(bm1^2*bm2^2))*m1*m2/s^3;
            % ddS2d2d2 = -(as/bm1)^2*(s.^2 + (as/bm1)^2*m1.^2)/s^3;
            
            ddSdM1dM1 =-(as^4*(bm2^2 - m2^2))/(bm1^2*bm2^2*s^3);
            ddSdM1dM2 =-(as^4*m1*m2)/(bm1^2*bm2^2*s^3);
            ddSdM2dM1 =-(as^4*m1*m2)/(bm1^2*bm2^2*s^3);
            ddSdM2dM2 =-(as^4*(bm1^2 - m1^2))/(bm1^2*bm2^2*s^3);
            
            % SecondGrad = [ddS1d1d1 ddS1d1d2
            %               ddS2d2d1 ddS2d2d2];
            
            SecondGrad = [ddSdM1dM1 ddSdM1dM2
                          ddSdM2dM1 ddSdM2dM2];
        end
        
        function [vecOut]= distance(obj,ps)
            %Gives vector to closest boundary point
            pboundary = obj.ClosestPoint(ps);
            vecOut = pboundary-ps;
        end
        
       function [pOut]= ClosestPoint(obj,p) %Gives point on closest boundary
            Np = size(p,1);
            pOut = zeros(size(p));
%             options = optimoptions('fmincon','SpecifyObjectiveGradient',true,'Algorithm','sqp')  % indicate gradient is provided
            
            for i = 1:Np
                Fmin = @(x) (p(i,1) - x(1)).^2 + (p(i,2) - x(2)).^2 + (p(i,3) - x(3)).^2;

%                 Fmin = @(x) obj.DistFunc(p(i,:),x);
                nonlcon = @obj.ConstraintEq;
                pOuttmp = fmincon(@(x) Fmin(x),p(i,:),[],[],[],[],[],[],nonlcon);%,options);                
                pOut(i,:) = pOuttmp';
            end
            
        end
         
        function [c,ceq] = ConstraintEq(obj,xlcl)
            c = [];
%             ceq = obj.F(x');            
            ceq = ((xlcl(1)./obj.Xscale).^2 + (xlcl(2)./obj.Yscale).^2 + (xlcl(3)./obj.Zscale).^2).^0.5 - 1;
        end
        
        function [f, g, H] = DistFunc(obj,p,x)
             i=1;
             f = (p(i,1) - x(1)).^2 + (p(i,3) - x(3)).^2 + (p(i,3) - x(3)).^2;
             g = [2*(p(i,1) - x(1));
                  2*(p(i,2) - x(2));
                  2*(p(i,3) - x(3))];
             H = [2 0 0; 0 2 0; 0 0 2];
              
            
        end
        
        function [Supdate] = FuncFelimActual(obj,p,update,slavedof)
            %ddFelimActual Second order actual derivitives for ellimination
            %  used for La Grange and CB2 Ellim approx
                       
            
            if slavedof == 1
                %x s
                s = p(1,1);
                m1 = p(1,2);
                m2 = p(1,3);
                a = obj.Xscale; 
                b = obj.Yscale;
                c = obj.Zscale;
                
            elseif slavedof == 2
                s = p(1,2);
                m1 = p(1,1);
                m2 = p(1,3);
                a = obj.Yscale; 
                b = obj.Xscale;
                c = obj.Zscale;
                
            elseif slavedof == 3
                s = p(1,3);
                m1 = p(1,1);
                m2 = p(1,2);
                a = obj.Zscale; 
                b = obj.Xscale;
                c = obj.Yscale;
                
                
            end
            m1 = m1 + update(1,1);
            m2 = m2 + update(2,1);
            
            Supdate1 = a*(1-(m1/b).^2 - (m2/c).^2)^0.5 - s;
            Supdate2 = -a*(1-(m1/b).^2 - (m2/c).^2)^0.5 - s;
            
            if abs(Supdate1) < abs(Supdate2)
                Supdate = Supdate1;
            else
                Supdate = Supdate2;
            end 
        end
        
        function xy = plot(obj)
            %plot Gives boudary plot points
            % This returns signed distance function for a given point
            
            x(:,1) = linspace(-obj.Xscale,obj.Xscale,1000);
            y1 = (-(x./obj.Xscale).^2 + 1).^0.5.*obj.Zscale;
            y2 = -(-(x./obj.Xscale).^2 + 1).^0.5*obj.Zscale;
            xy =[x, y1;
                flip(x), flip(y2)];
        end
        
        function [Grad] = tangent(obj,p)
            %dF Summary of this method goes here
            %  This returns the tangent at the point
            
            dRdu1(:,1) = (2*p(:,1))./obj.Xscale^2;
            dRdu2(:,1) = (2*p(:,2))./obj.Zscale^2;
            lclGrad = [dRdu1 dRdu2];
            Grad  = lclGrad./norm(lclGrad);
        end
        
    end
    
    methods (Access = private)
        
        function [F, G] = myoptimfunction(obj,f,g,t)
            %dF Summary of this method goes here
            %  This returnd the first order partial derivitves of the function
            F = f(t);
            G = g(t);
            
            
        end
        
        
    end
end

