function [p,tet, iters, updates, AverageLcLdes, MeshedAtIter] = CB_distmesh3d_mpc(fFunc,fH,h_max,bbox,pfix,endtol,TrussType,TrussS,scale,varargin)
%DISTMESH2D 2-D Mesh Generator using Distance Functions.
%   [P,T] = CB_distmesh2d_mpc(fFunc,fH,h_max,h_min,bbox,pfix,pmpc,endtol,varargin
%
%      P:         Node positions (Nx2)
%      T:         Triangle indices (NTx3)
%      fFunc:     Distance function d(x,y)
%      FH:        Scaled edge length function h(x,y)
%      h_max:     MaxEdge length(associated fH function)
%      h_min:     MaxEdge length(associated fH function)
%      BBOX:      Bounding box [xmin,ymin; xmax,ymax]
%      PFIX:      Fixed node positions (NFIXx2)
%      PMPC:      Any initiated MPC points
%      Endtol:    Termination tolerence for Newton step

%   See also: MESHDEMO2D, DISTMESHND, DELAUNAYN, TRIMESH.

%   distmesh2d.m v1.1
%   Copyright (C) 2004-2012 Per-Olof Persson. See COPYRIGHT.TXT for details.
tic

%% Set up parameters
ttol = 0.2; % Change tolerence for p movement
geps = 0.1;  % This the boudary limit for eliminating points and triangles
densityctrlfreq = 3; % destiny control frequency
destL = 1.6; % Length factor for length control on elements
UpdateLimiter = 20; % p update limiter This enforces stability for highly tensioned meshes
AllowNewts = 5; % Newtons teps allowed before remeshing
ploton  = true; %Plot during evaluations (see mesh updates

%% MPC Method selection STAYS THE SAME
MPCType =1;

if MPCType ==  1 %La grange
    MPC_lagrange = 1; % 1 use la grange MPC 0 use ellimination
    EllimACt = 0; % Actual 1 approx 0 for Elimination method only
elseif MPCType ==  2 %Ellim actual
    MPC_lagrange = 0; % 1 use la grange MPC 0 use ellimination
    EllimACt = 1; % Actual 1 approx 0 for Elimination method only
else %Ellim approx
    MPC_lagrange = 0; % 1 use la grange MPC 0 use ellimination
    EllimACt = 0; % Actual 1 approx 0 for Elimination method only
end

%% 1. Create initial distribution in bounding box (equilateral triangles) COMPLETED CONVERSION to 3d

% h02 = 0.8*(0.1*h_max + .9*h_min); %initial mesh seed edge length
h02 = h_max;


[x,y,z] = meshgrid(bbox(1,1):h02:bbox(2,1),bbox(1,2):h02*sqrt(3)/2:bbox(2,2),bbox(1,3):h02*sqrt(6)/3:bbox(2,3));
x(2:2:end,:,:) = x(2:2:end,:,:) + h02/2; % Shift even xz planes in x
y (:,:,2:2:end) = y(:,:,2:2:end) + h02*1/sqrt(3); % Shift even xy planes in x to create the tetrahedrals
p = [x(:),y(:),z(:)]; % List of all node coordinates

%% 2. Remove points outside the region, apply the rejection method STAYS THE SAME

p = p(fFunc.fd(p)<+geps.*fH(p),:);

% r0 = 1./feval(fH,p,varargin{:}).^2;                  % Probability to keep point
% p = p(rand(size(p,1),1)<r0./max(r0),:);              % Rejection method
pmpc = zeros(0,3);
if ~isempty(pfix), p = setdiff(p,pfix,'rows'); end     % Remove duplicated nodes
if ~isempty(pmpc), p = setdiff(p,pmpc,'rows'); end     % Remove duplicated nodes cb2

pfix = unique(pfix,'rows'); nfix = size(pfix,1);

%% Find unique mpc points from the input values (remove Fixed nodes)
pmpc = [pfix; pmpc];
[ ~, IA ,~] =   unique(pmpc,'rows','first');
IA = sort(IA,'ascend');
pmpc = pmpc(IA);
pmpc = pmpc(nfix+1:end,:); nmpc = size(pmpc,1);

%% Create enumeration types 0 free, 1 fixed, 2 mpc, 3 forcesnap
s = 4;
typeENUM.fixed = 0+s;
typeENUM.mpc = 1+s;
typeENUM.free = 2+s;
typeENUM.forcesnap = -1;

% Ensure unique fixed nodes
p = [pfix; pmpc; p];                                        % Prepend fix points
% p = unique(p,'rows','stable');

[~, IA, ~] = unique(p,'rows','first'); %keep unique points selecting first retains the Pfix, and PMC numbers
IA = sort(IA,'ascend');
p = p(IA,:);


FixMpcTypes = [typeENUM.fixed*ones(size(pfix,1),1); typeENUM.mpc*ones(size(pmpc,1),1)];
types = typeENUM.free*ones(size(p,1),1);
types(1:size(FixMpcTypes,1)) = FixMpcTypes;

% For first variable initialisation
counter = 0;
counter2 = 0;
pold = inf;
removedpcount = 0;
addedpcount = 0;
minqual = 0;
densitynewts = inf;
newtonsteps = inf;
ForcedRemesh = inf;
ForceNewts = false;
remesh = false;
updates = zeros(0,1);
tet = inf;
MeshedAtIter = zeros(0,1);
NewtCount = 1;
load('Mesh_intial.mat'); remesh =true;
while 1
    counter = counter +1;
    if counter == 10
        disp('100')
    end
    
    if ForceNewts ==  false
        %% 3. Retriangulation by the Delaunay algorithm
        if ((max(sqrt(sum((p-pold).^2,2))./feval(fH,p)) > ttol) && (newtonsteps > AllowNewts) && (minqual < 0.1)) || (ForcedRemesh >0)          % Any large movement?
            
            [p, In] = sortrows(p,'ascend');
            types = types(In,:);
            
            [~,IA,~] = unique(p,'rows','first');
            IA = sort(IA,'ascend'); % Keep the rows stable
            p = p(IA,:);
            types = types(IA);
            
            
            tetold = tet;
                                                      % Save current positions
            tet = delaunay(p);                                  % List of triangles
            
            pmid = (p(tet(:,1),:)+p(tet(:,2),:)+p(tet(:,3),:)+p(tet(:,4),:))/4;    % Compute triangle centroids
            tet = tet(fFunc.fd(pmid)<+(.05*fH(pmid)),:);         % Keep interior triangles only
            
            %% Find low quality triangle
%             quals = simpqual(p,tet);
%             lowquals = find(quals <= 0.1);
%             
%             if (size(lowquals,1) > 0) && (ForcedRemesh == 0)
%                 newps = zeros(0,2);
%                 numlowtris  = size(lowquals,1);
%                 newps = (p(tet(lowquals,1),:)+p(tet(lowquals,2),:)+p(tet(lowquals,3),:)+p(tet(lowquals,4),:))/4; % add mid points
%                 p = [p; newps];
%                 types = [types; ones(numlowtris,1)*typeENUM.free];
%                 tet = delaunay(p);
%                 
%                 pmid = (p(tet(:,1),:)+p(tet(:,2),:)+p(tet(:,3),:)+p(tet(:,4),:))/4;    % Compute triangle centroids
%                 tet = tet(fFunc.fd(pmid)<-(geps.*fH(pmid)),:);         % Keep interior triangles only
%             end
            
            pold = p;
            
            %% 4. Describe each bar by a unique pair of nodes
            bars = [tet(:,[1,2]);tet(:,[1,3]);tet(:,[1,4]);tet(:,[2,3]);tet(:,[2,4]);tet(:,[3,4])]; % Interior bars duplicated
            bars = sort(bars,2);
            bars = unique(bars,'rows'); %sorts and orders nodes
            
            faces = [tet(:,[1,2,3]);
                tet(:,[1,2,4]);
                tet(:,[1,3,4]);
                tet(:,[2,3,4]);];
            
            faces = sort(faces,2);
            [faces, ~, ic] = unique(faces,'rows');
            
            %  Add the triangles on the surface of the mesh to the MPC list to force a
            %  step to the edge of the solver
            SurfFace =  find(histc(ic,min(ic):max(ic)) ==  1); % find unique triangles (edge only faces)
            SurfPs = unique(reshape(faces(SurfFace,:),{},1)); % find surface ps
            
            % Find all points on the boudary of the mesh and set them to
            % forced snap (unique triangles are boundary)
            
            MPCsurfPs = setdiff(SurfPs,[find(types == typeENUM.mpc);find(types == typeENUM.fixed)]);
            types(MPCsurfPs) = typeENUM.forcesnap;
            
            %Find all edge Tets and collect there ps including interior ps for
            % Checking of MPC conditions
            edgeTetss = zeros(0,1);
            for EdgeP = 1:size(SurfPs,1)
                p2find = SurfPs(EdgeP);
                edgeTetss = [edgeTetss;
                    find(tet(:,1) == p2find);
                    find(tet(:,2) == p2find);
                    find(tet(:,3) == p2find);
                    find(tet(:,4) == p2find)];
            end
            
            
            if size(tetold,1)~= size(tet,1)
                MeshedAtIter(NewtCount,1) = 1;
            else
                if sum(any((tet-tetold) ~=0)) > 0
                    MeshedAtIter(NewtCount,1) = 1;
                end
            end
            
            edgeTetss = unique(edgeTetss);
            EdgeCheckPs = unique(reshape(tet(edgeTetss,:),{},1),'rows'); % Create a list of all the Ps that are in edge triangles this speeds up the search process
            EdgeCheckPs = setdiff(EdgeCheckPs,find(types == typeENUM.fixed));

            % 5. Graphical output of the current mesh
            if ploton ==  true
                MESHplotter3D(p,tet,MPCType)
                hold on;plot3(p(SurfPs,1),p(SurfPs,2),p(SurfPs,3),'k*');hold off
                legend('Element faces', 'Surface Fixed Nodes')
            end
            if ForcedRemesh > 0
                disp(['Forced ' num2str(ForcedRemesh)])
            else
                disp('auto remeshed')
            end
            
            remesh = true;
            ForcedRemesh = 0;% Cb2
        end
        
        %% Snap points to boundaries if required (creates MPCs reference)
        
        % Check only Ps in edge triangles
        fDs = fFunc.fd(p(EdgeCheckPs,:));
        snapPoints = EdgeCheckPs(find(fDs >-0.5*feval(fH,p(EdgeCheckPs,:))),1);
        snapPoints = unique([snapPoints; find(types == typeENUM.forcesnap)]);
        
        %Find mpc points that are no longer close to the boudary
        NotMpcs = setdiff(find(types == typeENUM.mpc),snapPoints);
        types(NotMpcs,1) = typeENUM.free;
        
        snapPoints = setdiff(snapPoints,find(types == typeENUM.fixed));
        %%   snapPoints = setdiff(snapPoints,find(types == type.mpc));  %Removes mpc points from snapping
        if size(snapPoints,1) > 0
            if size(find(types(snapPoints) ==  typeENUM.mpc),1) < size(snapPoints,1)
                types(snapPoints) = typeENUM.fixed;
                %cb2
                pSnap = p(snapPoints,:);
                dvecsnap = fFunc.distance(pSnap); % Shortest distance vec
                
                p(snapPoints,:) =  pSnap+ dvecsnap;
                plotvecstobound  = 0;
                if plotvecstobound ==  1 % Plot lines that are used for the snap
                    for ii = 1:size(snapPoints,1)
                        hold on
                        lclp = [pSnap(ii,:); (pSnap(ii,:)+ dvecsnap(ii,:))];
                        plot3(lclp(:,1),lclp(:,2),lclp(:,3),'r-')
                    end
                end                
                pold = inf;
                newtonsteps = inf;
                ForcedRemesh = 1;
                continue
            end
        end
        
        %% cb2 remove external points not part of a bar
        % Still to evaluate neccesity
        
        p2remove = setdiff(1:size(p,1),bars)';
        if size(p2remove,1) > 0
            p(p2remove,:) = [];
            types(p2remove,:) = [];
            removedpcount =  removedpcount + size(p2remove,1);
            pold = inf;
            newtonsteps = inf;
            ForcedRemesh = 2;
            continue
        end
        
    end
    %% 6. Create bar lengths
    
    if remesh ==  true || (mod(newtonsteps,AllowNewts) == 0)   % Recalculate bar lengths every allowed newton step, this allows for quadratic convergence
        barvec = p(bars(:,1),:) - p(bars(:,2),:);            % List of bar vectors
        Lc = sqrt(sum(barvec.^2,2));                          % L = current bar lengths
        hbars = feval(fH,(p(bars(:,1),:)+p(bars(:,2),:))/2,varargin{:}); % Actual desired lengths
%         L0 = hbars*sqrt(sum(L.^2)/sum(hbars.^2)); L0act = L0;    % L0 = Desired lengths scaled by sum of the squares
%         L0 = hbars*sqrt(sum(L.^2)/sum(hbars.^2)); L0act = hbars;
        Ldes = hbars; Ldes = hbars;
        disp('LENGTH CHANGED')
    end
    
    %% CB2   Density control - remove/average points that are too close together
    if ForceNewts ==  false
        %% short edge density control STAYS THE SAME
        if ((mod(counter,densityctrlfreq)==0)  && any(Ldes>destL*Lc)) || counter == 0 || any(Ldes>3*destL*Lc)
            index = find(Ldes>destL*Lc);
            ShrtLs = Lc(index,:);
            [~, InDx] = sort(ShrtLs,'ascend');
            shrtbars = bars(index,:);
            shrtbars = shrtbars(InDx,:);
            
            removedps = zeros(0,1);
            barnum = 0;
            
            while barnum < size(shrtbars,1)
                barnum =  barnum + 1;
                p1 = shrtbars(barnum,1);
                p2 = shrtbars(barnum,2);
                % Check if bar is to fixed node
                removedp = 0;
                if (types(p1,1) == typeENUM.fixed) && (types(p2,1) == typeENUM.fixed) % fixed fixed
                    % do nothing
                    continue
                elseif (types(p1,1) == typeENUM.fixed) && (types(p2,1) ~= typeENUM.fixed) % fixed & not fixed
                    % remove not fixed node
                    removedp = p2;
                elseif (types(p1,1) ~= typeENUM.fixed) && (types(p2,1) == typeENUM.fixed) % not fixed & fixed
                    % remove not fixed node
                    removedp = p1;
                elseif (types(p1,1) == typeENUM.mpc) && (types(p2,1) == typeENUM.mpc) % mpc &mpc
                    %take average point and make p1 as mpc and p2 removed
                    p(p1,:) = (p(p1,:)+p(p2,:))/2;
                    types(p1) = typeENUM.mpc;
                    removedp = p2;
                elseif (types(p1,1) == typeENUM.mpc) && (types(p2,1) ~= typeENUM.mpc) % mpc & not mpc
                    % leave mpc remove other point
                    removedp = p2;
                elseif (types(p1,1) ~= typeENUM.mpc) && (types(p2,1) == typeENUM.mpc) % not mpc & mpc
                    % leave mpc remove other point
                    removedp = p1;
                elseif (types(p1,1) == typeENUM.free) && (types(p2,1) == typeENUM.free) % free & free
                    %take average point and make p1 as free, p2 is removed
                    p(p1,:) = (p(p1,:)+p(p2,:))/2;
                    types(p1) = typeENUM.free;
                    removedp = p2;
                end
                                
                if (removedp ~= 0) %overload control if zero, must not be considered
                    removedps =[removedps;removedp];
                end
                
                removedpcount = removedpcount +1;
                %rows containing the removed point
                rowsCR = unique([find(shrtbars(barnum+1:end,1)==removedp);find(shrtbars(barnum+1:end,2)==removedp)])+ barnum;
                
                if size(rowsCR,1) > 0
                    shrtbars(rowsCR,:) = []; % remove rows containing the removed point
                end
            end
            if size(removedps,1)>0
                p(removedps,:) = [];
                types(removedps,:) = [];
                
                pold=inf;
                newtonsteps = inf;
                ForcedRemesh = 3;
                continue;
            end
            
        end
        
        %Long edge handleing simple Low effectiveness
        
        %     if ((mod(count,densityctrlfreq) == 0 ) || ((Forcedestiny ==  1)) ) && any(L>destL./2*L0act)
        %         LongBars = bars(L>destL./2*L0act,:);
        %
        %         newps = (p(LongBars(:,1),:)+p(LongBars(:,2),:))/2;
        %
        %         newtypes = type.free*ones(size(newps,1),1);
        %         p = [p; newps];
        %         types = [types; newtypes];
        %
        %         addedpcount = addedpcount+ size(LongBars,1);
        %
        %         pold = inf;
        %         newtonsteps = inf;
        %         ForcedRemesh = true;
        %         continue;
        %     end
        
        %% Long bars edge control (add points) STAYS THE SAME
        testdest = sum(Lc)/sum(Ldes);
        testdest = mean(Lc./Ldes);
        if testdest < 1.4 && (testdest > 0.7 || any(Ldes*destL*2 < Lc))
            if ((mod(counter,densityctrlfreq)==0 )  && any(Lc>destL*Ldes)) || any(Lc>3*destL*Ldes)%
                
                index = find(Lc>destL*Ldes);
                
                LongLs = Lc(index,:);
                LongL0acts = Ldes(index,:);
                LongBars = bars(index,:);
                
                [LongLs,InDx] = sort(LongLs,'descend');
                LongL0acts = LongL0acts(InDx,:);
                LongBars = LongBars(InDx,:);
                
                newps = zeros(0,1);
                newTypes = zeros(0,1);
                barnum = 0;
                
                while barnum < size(LongBars,1)
                    barnum = barnum + 1;
                    p1 = LongBars(barnum,1);
                    p2 = LongBars(barnum,2);
                    
                    NoPsToAdd = floor(LongLs(barnum,:)/(destL*LongL0acts(barnum,1)));
                    %                 NoPsToAdd = floor(LongLs(barnum,:)/(LongL0acts(barnum,1)));
                    AddedPs = p(p1,:) + 1/(NoPsToAdd+1)*[1:NoPsToAdd]'.*(p(p2,:)-p(p1,:));
                                        
                    %remove long bars that have an edge attached to this
                    AddedTypes = ones(NoPsToAdd,1)*typeENUM.free;
                    rowsCR = unique([find(LongBars(barnum+1:end,1)==LongBars(barnum,1));
                        find(LongBars(barnum+1:end,2)==LongBars(barnum,1));
                        find(LongBars(barnum+1:end,1)==LongBars(barnum,2));
                        find(LongBars(barnum+1:end,2)==LongBars(barnum,2))]);%+ barnum;
                    
                    if size(rowsCR,1) > 0
                        LongBars(rowsCR,:) = []; % remove rows containing the removed point
                        LongL0acts(rowsCR,:) = [];
                        LongLs(rowsCR,:) = [];
                    end
                    if size(AddedPs,1) ~= size(AddedTypes,1)
                        disp('failed to add types')
                        pause
                    end
                    newps = [newps; AddedPs];
                    newTypes = [newTypes; AddedTypes];
                end
                
                p = [p;newps];
                types = [types;newTypes];
                
                addedpcount = addedpcount+ size(newps,1);
                
                pold=inf;
                newtonsteps = inf;
                ForcedRemesh = 41;
                continue;
            end
        elseif testdest >= 1.4
            if ((mod(counter,densityctrlfreq)==0 )  && any(Lc>destL*Ldes)) || any(Lc>3*destL*Ldes)%
                
                side1 = sqrt(sum((p(tet(:,2),:)-p(tet(:,1),:)).^2,2));
                side2 = sqrt(sum((p(tet(:,3),:)-p(tet(:,1),:)).^2,2));
                side3 = sqrt(sum((p(tet(:,3),:)-p(tet(:,2),:)).^2,2));
                side4 = sqrt(sum((p(tet(:,4),:)-p(tet(:,1),:)).^2,2));
                side5 = sqrt(sum((p(tet(:,4),:)-p(tet(:,2),:)).^2,2));
                side6 = sqrt(sum((p(tet(:,4),:)-p(tet(:,3),:)).^2,2));
                
                tcircs = side1+side2+side3+side4+side5+side6;
                pmid = (p(tet(:,1),:)+p(tet(:,2),:)+p(tet(:,3))+p(tet(:,4),:))/4;    % Compute triangle centroids
                Largets = find(tcircs./6 > destL.*fH(pmid));
                
                newps = (p(tet(Largets,1),:)+p(tet(Largets,2),:)+p(tet(Largets,3))+p(tet(Largets,4),:))/4;
                
                if size(newps,1)>0
                    p = [p; newps];
                    types = [types; ones(size(newps,1),1)*typeENUM.free];
                    
                    pold=inf;
                    newtonsteps = inf;
                    ForcedRemesh = 42;
                    continue;
                end
            end
        end

        Forcedensity = 0;
    end
    %% CB2 Start solve code variable initialisation
    if remesh ==  true || counter == 1
        % number of mpc nodes lagrange
        MPCnums = find(types == typeENUM.mpc);
        nmpc = size(MPCnums,1);
        MPCdof = sort([MPCnums*3-2; MPCnums*3-1; MPCnums*3]);
        
        % find number of fixed nodes
        pnums = find(types == typeENUM.fixed);
        pdof = sort([pnums*3-2;pnums*3-1;pnums*3]);
        
        % find number of free nodes
        fnums = find(types == typeENUM.free);
        fdof = sort([fnums*3-2;fnums*3-1;fnums*3]);
        ndof = size(MPCdof,1)+size(pdof,1)+size(fdof,1); %truss contribution total dofs
        
        u = zeros(ndof,1);
        
        lambda = 0;
        
        remesh = false;
        newtonsteps = 0;
    end
    
    
    
    %% Truss Consistant tangent contributions
    F = zeros(ndof,1);
    DFDU = zeros(ndof,ndof);
    
    for i = 1:size(bars,1)
        dof = [bars(i,1)*3-2, bars(i,1)*3-1,bars(i,1)*3-0, bars(i,2)*3-2, bars(i,2)*3-1, bars(i,2)*3-0]';
        L0bar = Ldes(i,1); %desired length of bar
        uloc = u(dof);
        x = [pold(bars(i,1),:), pold(bars(i,2),:)]';
        [dFdu,broken,Ft] = CB_3dtruss_CT(uloc,x,L0bar,TrussType,TrussS); % CB_2dtruss_CT has 3x methods for Bar contribution.
        
        if broken ==  true
            break
        end
        
        DFDU(dof,dof) = DFDU(dof,dof) + dFdu;
        
        F(dof) = F(dof) + Ft;
    end
    
    
    %% MPC methods contributions
    % These mehtods can be selected at the input parameters at the top of this script
    % La grange and Ellimination methids have seperate newton step updates
    
    if 1 == MPC_lagrange %% Lagrange for circle
        mpccount = 0;
        DRmpcdUc = sparse(nmpc,ndof);
        DDRmpcddUc = zeros(ndof,ndof);
        
        if newtonsteps ==  0 %just remeshed
            lambda = zeros(nmpc,1); %only used for la grange
            ldof = ((ndof+1):(ndof+nmpc))';
        end
        
        for i = MPCnums'
            mpccount = mpccount +1;
            
            dof = i*3-2:i*3;
            
            pcur = pold(i,:);
            uloc = u(dof);
            
            Point2B = pcur + uloc';
            
            DRmpcdUc(mpccount,dof) = fFunc.dF(Point2B);
            
            ddRddu = fFunc.ddF(Point2B);
            
            DDRmpcddUc(dof,dof) = DDRmpcddUc(dof,dof)+lambda(mpccount,1)*ddRddu;
        end
        
        K_global = [(DFDU+DDRmpcddUc) DRmpcdUc'
            DRmpcdUc sparse(nmpc,nmpc)];
        
        RHS = sparse(ndof+nmpc,1);
        RHS(pdof,1) = F(pdof,1);
        RHS(fdof,1) = F(fdof,1);
        
        RHS(MPCdof,1) = F(MPCdof,1)+(DRmpcdUc(:,MPCdof))'*lambda;
        
        %solve dof
        solvedof = [fdof;MPCdof;ldof];
        
        % Newton update
        Supdate = zeros(ndof+nmpc,1);
        Supdate(solvedof) = -K_global(solvedof,solvedof)\RHS(solvedof);
        
        % Reshape to same as p
        update = Supdate(1:ndof,1);
        delta_lam = Supdate(ldof);
        
        %     %Min Norm update attempt
        P_update = [update(1:3:end,1), update(2:3:end,1),update(3:3:end,1)];
        
        % Determine proportion of update to be used
        % if any bar wants to change by more than half its length, then reduce
        % update size
        
        barvecupdate = P_update(bars(:,1),:)-P_update (bars(:,2),:);
        barchange = sqrt(sum(barvecupdate.^2,2));
        change = abs(barchange./Lc);
        maxchange = max(change);
        
        if maxchange > UpdateLimiter
            delta = UpdateLimiter/maxchange;
        else
            delta = 1;
        end
        
        lambda = lambda + delta*delta_lam;
        u = u + delta*update;
        p = pold+[u(1:3:end),u(2:3:end),u(3:3:end)];
        
    else
        %% Elimination methods
        %Selection is done by the slave selection as the partial
        %derivitive with the greatest constribution.
        
        Slavedofs_temp = sparse(ndof,1);
        Masterdofs_temp = sparse(ndof,1);
        Pmat_temp = sparse(ndof,ndof);
        dPdm_mat_temp = sparse(ndof,ndof);
        mpccount = 0;
        for i = MPCnums'
            mpccount = mpccount +1;
            
            dof = i*3-2:i*3;
            
            pcur = pold(i,:);
            uloc = u(dof);
            
            Point2B = pcur + uloc';
            
            dRdu = fFunc.dF(Point2B);
            
            dRdu1 = dRdu(1,1);
            dRdu2 = dRdu(1,2);
            dRdu3 = dRdu(1,3);
            
            if 0 ==  EllimACt % CB2 second partial dirivitive approximation
                ddRddu = fFunc.ddF(Point2B);
                
                ddRdu1du1 = ddRddu(1,1);
                ddRdu1du2 = ddRddu(1,2);
                ddRdu1du3 = ddRddu(1,3);
                ddRdu2du1 = ddRddu(2,1);
                ddRdu2du2 = ddRddu(2,2);
                ddRdu2du3 = ddRddu(2,3);
                ddRdu3du1 = ddRddu(3,1);
                ddRdu3du2 = ddRddu(3,2);
                ddRdu3du3 = ddRddu(3,3);
                
                if (abs(dRdu1) >=  abs(dRdu2)) && (abs(dRdu1) >=  abs(dRdu3))
                    MasterDof = dof(1,[2 3]); %master is dof2 dof3
                    SlaveDof = dof(1,1);
                    Pmatlocal = [-dRdu2/dRdu1, -dRdu3/dRdu1];
                    
                    dd1d2d2 = -(ddRdu2du2*dRdu1 - dRdu2*ddRdu1du2)/(dRdu1^2);
                    dd1d2d3 = -(ddRdu2du3*dRdu1 - dRdu2*ddRdu1du3)/(dRdu1^2);
                    dd1d3d2 = -(ddRdu3du2*dRdu1 - dRdu3*ddRdu1du2)/(dRdu1^2);
                    dd1d3d3 = -(ddRdu3du3*dRdu1 - dRdu3*ddRdu1du3)/(dRdu1^2);
                    dPmatlocal = [dd1d2d2 dd1d2d3;
                        dd1d3d2 dd1d3d3];
                elseif (abs(dRdu2) >=  abs(dRdu1)) && (abs(dRdu2) >=  abs(dRdu3))
                    MasterDof = dof(1,[1 3]); %master is dof1 dof3
                    SlaveDof = dof(1,2);
                    Pmatlocal = [-dRdu1/dRdu2, -dRdu3/dRdu2];
                    
                    dd2d1d1 = -(ddRdu1du1*dRdu2 - dRdu1*ddRdu2du1)/(dRdu2^2);
                    dd2d1d3 = -(ddRdu1du3*dRdu2 - dRdu1*ddRdu2du3)/(dRdu2^2);
                    dd2d3d1 = -(ddRdu3du1*dRdu2 - dRdu3*ddRdu2du1)/(dRdu2^2);
                    dd2d3d3 = -(ddRdu3du3*dRdu2 - dRdu3*ddRdu2du3)/(dRdu2^2);
                    dPmatlocal = [dd2d1d1 dd2d1d3;
                        dd2d3d1 dd2d3d3];
                elseif (abs(dRdu3) >=  abs(dRdu1)) && (abs(dRdu3) >=  abs(dRdu2))
                    MasterDof = dof(1,[1 2]); %master is dof1 dof3
                    SlaveDof = dof(1,3);
                    Pmatlocal = [-dRdu1/dRdu3, -dRdu2/dRdu3];
                    
                    dd3d1d1 = -(ddRdu1du1*dRdu3 - dRdu1*ddRdu3du1)/(dRdu3^2);
                    dd3d1d2 = -(ddRdu1du2*dRdu3 - dRdu1*ddRdu3du2)/(dRdu3^2);
                    dd3d2d1 = -(ddRdu2du1*dRdu3 - dRdu2*ddRdu3du1)/(dRdu3^2);
                    dd3d2d2 = -(ddRdu2du2*dRdu3 - dRdu2*ddRdu3du2)/(dRdu3^2);
                    dPmatlocal = [dd3d1d1 dd3d1d2;
                        dd3d2d1 dd3d2d2];
                    
                end
            else %actual second derivitive
                if (abs(dRdu1) >=  abs(dRdu2)) && (abs(dRdu1) >=  abs(dRdu3))
                    sdof = 1;
                elseif (abs(dRdu2) >=  abs(dRdu1)) && (abs(dRdu2) >=  abs(dRdu3))
                    sdof = 2;
                elseif (abs(dRdu3) >=  abs(dRdu1)) && (abs(dRdu3) >=  abs(dRdu2))
                    sdof = 3;
                end
                SlaveDof = dof(1,sdof);
                MasterDof= dof(1, setdiff(1:3,sdof)); %master is dof2
                Pmatlocal = fFunc.dFelimActual(Point2B,sdof);
                dPmatlocal = fFunc.ddFelimActual(Point2B,sdof);
            end
            
            
            Slavedofs_temp(SlaveDof,1) = 1;
            Masterdofs_temp(MasterDof,1) = 1;
            Pmat_temp(SlaveDof,MasterDof) = Pmatlocal;
            dPdm_mat_temp(MasterDof,MasterDof) = dPmatlocal*F(SlaveDof,1);
        end
        
        SlaveDofs = find(Slavedofs_temp == 1);
        MasterDofs = find(Masterdofs_temp == 1);
        
        %Solve degrees of freedom
        solvedof = [fdof;
            MasterDofs];
        
        Pmat = Pmat_temp(SlaveDofs,MasterDofs);
        dPdm_mat_term2 = dPdm_mat_temp(MasterDofs,MasterDofs);
        
        RHS_f = F(fdof,1);
        RHS_ms = F(MasterDofs,1) + Pmat'*F(SlaveDofs,1);
        RHS = [RHS_f;
            RHS_ms];
        
        Kff = DFDU(fdof,fdof);
        Kfm = DFDU(fdof,MasterDofs) + DFDU(fdof,SlaveDofs)*Pmat;
        Kmf = DFDU(MasterDofs,fdof) + Pmat'*DFDU(SlaveDofs,fdof);
        Kmm = DFDU(MasterDofs,MasterDofs) + dPdm_mat_term2 +  Pmat'*DFDU(SlaveDofs,MasterDofs) + DFDU(MasterDofs,SlaveDofs)*Pmat + Pmat'*DFDU(SlaveDofs,SlaveDofs)*Pmat;
        
        %Elmination Method mpcs K_global
        K_global = [Kff Kfm;
            Kmf Kmm];
        
        % Newton update
        Supdate = zeros(ndof,1);
        Supdate(solvedof,1) = -K_global\RHS;%% Lagrange for circle
        Supdate(SlaveDofs,1) = Pmat*Supdate(MasterDofs,1);
        
        % Reshape to same as p
        update = Supdate(1:ndof,1);%% Lagrange for circle
        P_update = [update(1:3:end,1), update(2:3:end,1), update(3:3:end,1)];
        
        % Determine proportion of update to be used
        % if any bar wants to change by more than half its length, then reduce
        % update size
        
        barvecupdate = P_update(bars(:,1),:)-P_update (bars(:,2),:);
        barchange = sqrt(sum(barvecupdate.^2,2));
        change = abs(barchange./Lc);
        maxchange = max(change);
        
        if maxchange > UpdateLimiter
            delta = UpdateLimiter/maxchange;
        else
            delta = 1;
        end
        
        u = u + delta*update;
        p = pold+[u(1:3:end),u(2:3:end),u(3:3:end)];
        
    end
    
    newtonsteps = newtonsteps +1;
    
    %% end CB2
    
    % 8. Termination criterion:
    if ploton ==  true
        %PlotMPCs points around edge
        hold on
        plot3(p(MPCnums,1),p(MPCnums,2),p(MPCnums,3),'r.')
        plot3(p(pnums,1),p(pnums,2),p(pnums,3),'k*')
        drawnow
    end
    %     axis([-3 3 -3 3])
    axis equal
    hold off
    
    %    if max(abs(update)) < endtol
    qual = simpqual(p,tet,1);
    minqual = min(qual);
    NormUpdate = norm(update)
    updates(size(updates,1)+1,1) = NormUpdate;
        if NormUpdate < 1
            AllowNewts = 5;
            densityctrlfreq = 5;
        end
    
    
    %     if minqual > 0.6
    %         ForceNewts = true;
    %         types(find(types == typeENUM.mpc)) = typeENUM.fixed;
    %         %AllowNewts = 3;
    %         densityctrlfreq = 3;
    %     else
    %         %AllowNewts = 3;
    %         densityctrlfreq = 5;
    %     end
    
    counter2 = counter2 +1;
    if (NormUpdate < endtol)
        break;
    end
    if NewtCount>2
        ForceNewts = true;
    end
    if NormUpdate > 2000 || isnan(NormUpdate)
        break;
    end
    if counter > 50
        break;
    end
    NewtCount = NewtCount+1;
    %     if max(sqrt(sum(deltat*Ftot(d<-geps,:).^2,2))/h0)<dptol, break; end
end

MESHplotter3D(p,tet,MPCType)

minqual = min(qual)

npsend = size(p,1)
removedpcount
addedpcount
iters  = counter

lr = Lc./Ldes;
AverageLcLdes = mean(Lc./Ldes);

minls = min(lr)
maxls = max(lr)
avelr = mean(lr)
time = toc

% pause

% Clean up and plot final mesh
% % % [p,t] = fixmesh(p,t);
% % % simpplot(p,t)
end

function [a] = MESHplotter3D(p,tet,fignum)
figure(fignum); cla
% simpplot(p,tet,'p(:,2)>0');
% title(['Retriangulation #'])
%         drawnow
t = unique(sort([tet(:,[1,2,3]);tet(:,[1,2,4]);tet(:,[1,3,4]);tet(:,[2,3,4]);],2),'rows');
cla;
s1 = patch('vertices',p,'faces',t,'edgecol','k','facecol',[.8,.9,1]);
% s1.FaceAlpha = 0.00;
axis equal

% draw low quality elements
qual = simpqual(p,tet,1);
AD = find(qual<.2);

t = unique(sort([tet(AD,[1,2,3]);tet(AD,[1,2,4]);tet(AD,[1,3,4]);tet(AD,[2,3,4]);],2),'rows');

% hold on
% s2 = patch('vertices',p,'faces',t,'edgecol','k','facecol','r'); drawnow;
% s2.FaceAlpha = 0.3;
% hold off
end


